from __future__ import unicode_literals

import datetime
import math
import pickle
import time

import pytest
from django.core.cache.backends.locmem import LocMemCache
from django.test.client import RequestFactory
from django.urls import reverse
from freezegun import freeze_time
from mock import patch
from rest_framework import status

from drf_jwt_2fa.throttling import CodeTokenThrottler

from .test_endpoints import get_code_token
from .utils import (
    OverrideJwt2faSettings, get_api_client, get_verification_code_from_mailbox)


@pytest.mark.parametrize('num', [None, 1, 42, 938383])
@pytest.mark.parametrize('duration_str, duration_seconds', [
    ('s', 1), ('7s', 7), ('m', 60), ('10m', 600), ('h', 3600), ('d', 86400)])
def test_code_token_throttler_parse_rate(num, duration_str, duration_seconds):
    rate_string = str(num) + '/' + duration_str if num else None
    expected_result = (num, duration_seconds) if num else (None, None)
    throttler = CodeTokenThrottler()
    throttler.parse_rate(rate_string) == expected_result


def get_code_token_throttler(cache):
    throttler = CodeTokenThrottler()
    throttler.timer = time.time
    throttler.cache = cache
    return throttler


@OverrideJwt2faSettings({'CODE_TOKEN_THROTTLE_RATE': '2/10s'})
def test_code_token_throttler():
    with patch('drf_jwt_2fa.throttling.sha1_string') as mocked_sha1_string:
        mocked_sha1_string.side_effect = lambda x: 'SHA1({})'.format(x)
        check_code_token_throttler(RequestFactory())


def check_code_token_throttler(rf):
    with freeze_time('2020-01-02 13:00:00') as frozen_datetime:
        cache = LocMemCache('test_cache', {})
        cache.clear()
        throttler = get_code_token_throttler(cache)
        request = rf.get('/')

        assert inspect_cache(cache) == {}

        # First two requests should be allowed
        assert throttler.allow_request(request, None) is True
        assert inspect_cache(cache) == {
            ':1:drf_jwt_2fa-tc-SHA1(127.0.0.1)': [1577970000.0],
        }

        frozen_datetime.tick(delta=datetime.timedelta(seconds=1))
        throttler = get_code_token_throttler(cache)
        assert throttler.allow_request(request, None) is True
        assert inspect_cache(cache) == {
            ':1:drf_jwt_2fa-tc-SHA1(127.0.0.1)': [1577970001.0, 1577970000.0],
        }

        # Third request should be throttled
        throttler = get_code_token_throttler(cache)
        frozen_datetime.tick(delta=datetime.timedelta(seconds=1))
        assert throttler.allow_request(request, None) is False
        assert inspect_cache(cache) == {
            ':1:drf_jwt_2fa-tc-SHA1(127.0.0.1)': [1577970001.0, 1577970000.0],
        }

        # After 8s more, total 11s have passed and a new request should
        # be allowed
        throttler = get_code_token_throttler(cache)
        frozen_datetime.tick(delta=datetime.timedelta(seconds=8))
        assert throttler.allow_request(request, None) is True
        assert inspect_cache(cache) == {
            ':1:drf_jwt_2fa-tc-SHA1(127.0.0.1)': [1577970010.0, 1577970001.0],
        }


def inspect_cache(cache):
    return {
        key: pickle.loads(pickled_value)
        for (key, pickled_value) in cache._cache.items()}


@pytest.mark.django_db
def test_code_token_throttling():
    with freeze_time('2020-01-02 13:00:00') as frozen_datetime:
        assert time.time() == 1577970000.0

        code_token1 = get_code_token()
        code1 = get_verification_code_from_mailbox()

        code_token2 = get_code_token()
        code2 = get_verification_code_from_mailbox()

        incorrect_codes = [
            code for code in [
                '1234567', '2345678', '3456789', '4567890', '5678901',
                '6789012', '7890123', '8901234', '9012345', '0123456']
            if code not in {code1, code2}]
        client = get_api_client()

        def attempt(code_token, code):
            return client.post(
                reverse('auth'),
                data={'code_token': code_token, 'code': code})

        # Note: Default throttle wait time is 2 seconds

        # Try 1 on code_token1
        frozen_datetime.tick(delta=datetime.timedelta(seconds=1))
        assert time.time() == 1577970001.0
        result0 = attempt(code_token1, incorrect_codes[0])
        assert result0.data == {
            'detail': 'Incorrect authentication credentials.'}
        assert result0.status_code == status.HTTP_403_FORBIDDEN

        # Try 2 on code_token1, after 0.5s.  Should be throttled
        frozen_datetime.tick(delta=datetime.timedelta(seconds=0.5))
        assert time.time() == 1577970001.5
        result1 = attempt(code_token1, incorrect_codes[1])
        assert result1.data == {
            'detail': (
                'Request was throttled. Expected available in {} seconds.'
                .format(math.ceil(2)))
        }
        assert result1.status_code == status.HTTP_429_TOO_MANY_REQUESTS

        # Try 3 on code_token1, after 1.75s.  Should NOT be throttled,
        # since 2.25s have passed since the last non-throttled (Try 1)
        frozen_datetime.tick(delta=datetime.timedelta(seconds=1.75))
        assert time.time() == 1577970003.25
        result2 = attempt(code_token1, incorrect_codes[2])
        assert result2.data == {
            'detail': 'Incorrect authentication credentials.'}
        assert result2.status_code == status.HTTP_403_FORBIDDEN

        # Try 4 on code_token1, after 0.75s.  Should be throttled.
        frozen_datetime.tick(delta=datetime.timedelta(seconds=0.75))
        assert time.time() == 1577970004.0
        result3 = attempt(code_token1, incorrect_codes[3])
        assert result3.status_code == status.HTTP_429_TOO_MANY_REQUESTS

        # Try with a different code token.  Should NOT be throttled.
        assert time.time() == 1577970004.0
        result4 = attempt(code_token2, incorrect_codes[4])
        assert result4.data == {
            'detail': 'Incorrect authentication credentials.'}
        assert result4.status_code == status.HTTP_403_FORBIDDEN

        # Try without a code token.  Should NOT be throttled.
        assert time.time() == 1577970004.0
        result4 = attempt('', 'abc')
        assert result4.data == {'code_token': ['This field may not be blank.']}
        assert result4.status_code == status.HTTP_400_BAD_REQUEST

        # Try 5 on code_token1, after 1.5s.  Should NOT be throttled
        # since 2.25s have passed since the last non-throttled (Try 3)
        frozen_datetime.tick(delta=datetime.timedelta(seconds=1.5))
        assert time.time() == 1577970005.5
        result5 = attempt(code_token1, incorrect_codes[5])
        assert result5.data == {
            'detail': 'Incorrect authentication credentials.'}
        assert result5.status_code == status.HTTP_403_FORBIDDEN



import datetime

from django.conf import settings
from rest_framework.settings import APISettings

from .utils import hash_string

USER_SETTINGS = getattr(settings, 'JWT2FA_AUTH', None)

DEFAULTS = {
    # Length of the verification code (digits)
    'CODE_LENGTH': 7,

    # Characters used in the verification code
    'CODE_CHARACTERS': '0123456789',

    # Secret key to use for signing the Code Tokens
    'CODE_TOKEN_SECRET_KEY': hash_string('2fa-code-' + settings.SECRET_KEY),

    # Secret string to extend the verification code with
    'CODE_EXTENSION_SECRET': hash_string('2fa-ext-' + settings.SECRET_KEY),

    # How long the code token is valid
    'CODE_EXPIRATION_TIME': datetime.timedelta(minutes=5),

    # Throttle limit for code token requests from same IP
    'CODE_TOKEN_THROTTLE_RATE': '12/3h',

    # How much time must pass between verification attempts, i.e. to
    # request authentication token with a with the same code token and a
    # verification code
    'AUTH_TOKEN_RETRY_WAIT_TIME': datetime.timedelta(seconds=2),

    # Function that sends the verification code to the user
    'CODE_SENDER': 'drf_jwt_2fa.sending.send_verification_code_via_email',

    # From Address used by the e-mail sender
    'EMAIL_SENDER_FROM_ADDRESS': settings.DEFAULT_FROM_EMAIL,

    # Set to this to a (translated) string to override the default
    # message subject of the e-mail sender
    'EMAIL_SENDER_SUBJECT_OVERRIDE': None,

    # Set to this to a (translated) string to override the default
    # message body of the e-mail sender
    'EMAIL_SENDER_BODY_OVERRIDE': None,
}

IMPORT_STRINGS = [
    'CODE_SENDER',
]

api_settings = APISettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS)



import time

import jwt
from django.contrib.auth import hashers as django_hashers
from django.utils.crypto import get_random_string
from django.utils.translation import ugettext as _
from rest_framework import exceptions

from .exceptions import VerificationCodeSendingFailed
from .sending import CodeSendingFailed, send_verification_code
from .settings import api_settings


class CodeTokenManager(object):
    jwt_algorithm = 'HS256'

    @property
    def code_length(self):
        return api_settings.CODE_LENGTH

    @property
    def code_chars(self):
        return api_settings.CODE_CHARACTERS

    def create_code_token(self, user):
        """
        Create a code token and send a new verification code.

        Create a new code token for given user with a new randomly
        generated verification code.  The code token is returned and the
        verification code is sent via another channel (e.g. e-mail).

        :type user: django.contrib.auth.models.AbstractBaseUser
        :rtype: str
        """
        code = self.generate_verification_code()
        payload = self.get_token_payload(user, code)
        try:
            self.send_verification_code(user, code)
        except CodeSendingFailed as error:
            raise VerificationCodeSendingFailed(error)
        return self.encode_token(payload)

    def check_code_token_and_code(self, token, code):
        """
        Check code token and related verification code.

        Check integrity of the given code token and check that the
        verification code is correct for the given token.  Return
        username of the verified user, if both are OK, or raise a
        validation error otherwise.

        :type token: str
        :param token: Code token to check
        :type code: str
        :param code: Verification code to check against the token
        :rtype: str
        :return: Username of the verified user
        """
        payload = self.decode_token(token)
        hashed_code = payload.get('vch')
        nonce = payload.get('vcn')
        if not self.is_verification_code_ok(code, nonce, hashed_code):
            raise exceptions.AuthenticationFailed()
        return payload.get('usr')

    def generate_verification_code(self):
        return get_random_string(self.code_length, self.code_chars)

    def get_token_payload(self, user, code):
        """
        Get code token for given user and verification code.

        :type user: django.contrib.auth.models.AbstractBaseUser
        :type code: str
        :rtype: dict[str, str|int]
        """
        now = int(time.time())
        expiration_time = api_settings.CODE_EXPIRATION_TIME
        expiration_seconds = int(expiration_time.total_seconds())
        (hashed_code, nonce) = self.hash_verification_code(code)
        return {
            'usr': user.get_username(),
            'vch': hashed_code,  # Verification Code Hash
            'vcn': nonce,  # Verification Code Nonce
            'iat': now,
            'exp': now + expiration_seconds,
        }

    def send_verification_code(self, user, code):
        send_verification_code(user, code)

    def encode_token(self, payload):
        key = api_settings.CODE_TOKEN_SECRET_KEY
        jwt_data = jwt.encode(payload, key, self.jwt_algorithm)
        return jwt_data.decode('utf-8')

    def decode_token(self, token):
        try:
            return jwt.decode(
                jwt=token,
                key=api_settings.CODE_TOKEN_SECRET_KEY,
                verify=True,
                algorithms=[self.jwt_algorithm])
        except jwt.ExpiredSignature:
            raise exceptions.PermissionDenied(_("Signature has expired."))
        except jwt.DecodeError:
            raise exceptions.AuthenticationFailed()

    def hash_verification_code(self, code):
        nonce = get_random_string(length=10)
        extended_code = self.extend_code(code, nonce)
        hashed_code = django_hashers.make_password(extended_code)
        return (hashed_code, nonce)

    def is_verification_code_ok(self, code, nonce, hashed_code):
        extended_code = self.extend_code(code, nonce)
        return django_hashers.check_password(extended_code, hashed_code)

    def extend_code(self, code, nonce):
        extension = api_settings.CODE_EXTENSION_SECRET
        return code + nonce + extension



import base64
import json
import time

import jwt
import six
from django.conf import settings
from django.core import mail
from django.test import override_settings
from django.test.utils import TestContextDecorator
from rest_framework.test import APIClient
from rest_framework_jwt.settings import api_settings as jwt_settings

from drf_jwt_2fa.settings import api_settings

str = six.text_type  # Make str be unicode on Python 2 too


def get_api_client():
    api_client = APIClient()
    api_client.default_format = 'json'
    return api_client


def check_auth_token(token, username='testuser', email='testuser@localhost'):
    payload = check_token_basics_and_get_payload(token)
    assert sorted(payload.keys()) == ['email', 'exp', 'user_id', 'username']
    assert isinstance(payload['user_id'], int)
    assert isinstance(payload['username'], str)
    assert isinstance(payload['email'], str)
    assert isinstance(payload['exp'], int)
    assert payload['exp'] > time.time()
    assert payload['username'] == username
    assert payload['email'] == email
    key = jwt_settings.JWT_SECRET_KEY
    jwt.decode(token, key, algorithms=['HS256'])


def check_code_token(token, username='testuser', verify=True):
    payload = check_token_basics_and_get_payload(token)
    assert sorted(payload.keys()) == ['exp', 'iat', 'usr', 'vch', 'vcn']
    assert isinstance(payload['exp'], int)
    assert isinstance(payload['iat'], int)
    assert isinstance(payload['usr'], str)
    assert isinstance(payload['vch'], str)
    assert isinstance(payload['vcn'], str)
    assert payload['exp'] > time.time()
    assert payload['iat'] <= time.time()
    assert payload['usr'] == username
    assert payload['vch'].startswith('pbkdf2_sha256$')
    assert len(payload['vcn']) == 10
    if verify:
        key = api_settings.CODE_TOKEN_SECRET_KEY
        jwt.decode(token, key, algorithms=['HS256'])


def check_token_basics_and_get_payload(token):
    assert token
    assert len(token.split('.')) == 3
    (header_part, payload_part, signature_part) = token.split('.')
    assert decode_jwt_part(header_part) == {'alg': 'HS256', 'typ': 'JWT'}
    return decode_jwt_part(payload_part)


def get_verification_code_from_mailbox():
    return mail.outbox[-1].subject.split(':')[0]


def decode_jwt_part(part):
    return json.loads(b64decode(part).decode('utf-8'))


def b64decode(encoded):
    padding = (4 - len(encoded) % 4) * '='
    return base64.b64decode(encoded + padding)


def encode_jwt_part(data):
    json_data = json.dumps(data, separators=(',', ':')).encode('utf-8')
    return base64.b64encode(json_data).decode('ascii').strip('=')


class OverrideJwt2faSettings(TestContextDecorator):
    def __init__(self, values):
        super(OverrideJwt2faSettings, self).__init__()
        self.override_settings = override_settings(JWT2FA_AUTH=values)

    def enable(self):
        self.override_settings.enable()
        api_settings.reload()
        api_settings._user_settings = settings.JWT2FA_AUTH

    def disable(self):
        self.override_settings.disable()
        api_settings.reload()



import time

from django.core.cache import cache as default_cache
from rest_framework import throttling

from .settings import api_settings
from .utils import sha1_string


class CodeTokenThrottler(throttling.SimpleRateThrottle):
    cache_key_template = 'drf_jwt_2fa-tc-{ident_hash}'

    @property
    def rate(self):
        return api_settings.CODE_TOKEN_THROTTLE_RATE

    def parse_rate(self, rate):
        if not rate:
            return (None, None)
        (num_requests_str, period_str) = rate.split('/')
        period_num = int(period_str[:-1] or '1')
        period_unit = {'s': 1, 'm': 60, 'h': 3600, 'd': 86400}[period_str[-1]]
        return (int(num_requests_str), period_num * period_unit)

    def get_cache_key(self, request, view):
        return self.cache_key_template.format(
            ident_hash=sha1_string(self.get_ident(request)))


class AuthTokenThrottler(throttling.BaseThrottle):
    cache = default_cache
    cache_key_template = 'drf_jwt_2fa-ta-{code_token_hash}'

    def allow_request(self, request, view):
        key = self.get_cache_key(request, view)
        if not key:
            return True
        now = time.time()
        next_allowed = self.cache.get(key)
        if next_allowed and next_allowed > now:
            self.wait_time = next_allowed - now
            return False
        self.cache.set(key, now + self.retry_wait_seconds)
        return True

    def wait(self):
        return self.wait_time

    def get_cache_key(self, request, view):
        code_token = request.data.get('code_token')
        return self.cache_key_template.format(
            code_token_hash=sha1_string(code_token)) if code_token else None

    @property
    def retry_wait_seconds(self):
        return api_settings.AUTH_TOKEN_RETRY_WAIT_TIME.total_seconds()



from django.contrib.auth import authenticate, get_user_model
from rest_framework import exceptions, serializers
from rest_framework_jwt.compat import Serializer as JwtSerializer
from rest_framework_jwt.serializers import PasswordField
from rest_framework_jwt.settings import api_settings as jwt_settings

from .token_manager import CodeTokenManager
from .utils import check_user_validity

jwt_encode_handler = jwt_settings.JWT_ENCODE_HANDLER
jwt_payload_handler = jwt_settings.JWT_PAYLOAD_HANDLER


class Jwt2faSerializer(JwtSerializer):
    token_manager_class = CodeTokenManager

    def __init__(self, *args, **kwargs):
        super(Jwt2faSerializer, self).__init__(*args, **kwargs)
        self.token_manager = self.token_manager_class()

    def validate(self, attrs):
        validated_attrs = super(Jwt2faSerializer, self).validate(attrs)
        user = self._authenticate(validated_attrs)
        return {
            'token': self._create_token(user),
            'user': user,
        }


class CodeTokenSerializer(Jwt2faSerializer):
    username = serializers.CharField(required=True)
    password = PasswordField(write_only=True, required=True)

    def _authenticate(self, attrs):
        credentials = {
            'username': attrs.get('username'),
            'password': attrs.get('password'),
        }
        user = authenticate(**credentials)
        if not user:
            raise exceptions.AuthenticationFailed()
        check_user_validity(user)
        return user

    def _create_token(self, user):
        return self.token_manager.create_code_token(user)


class AuthTokenSerializer(Jwt2faSerializer):
    code_token = serializers.CharField(required=True)
    code = PasswordField(write_only=True, required=True)

    def _authenticate(self, attrs):
        code_token = attrs.get('code_token')
        code = attrs.get('code')
        username = self._check_code_token_and_code(code_token, code)
        user = self._get_user(username)
        return user

    def _check_code_token_and_code(self, code_token, code):
        return self.token_manager.check_code_token_and_code(code_token, code)

    def _get_user(self, username):
        user_model = get_user_model()
        try:
            user = user_model.objects.get_by_natural_key(username)
        except user_model.DoesNotExist:
            raise exceptions.AuthenticationFailed()
        check_user_validity(user)
        return user

    def _create_token(self, user):
        payload = jwt_payload_handler(user)
        return jwt_encode_handler(payload)



import base64
import hashlib

from rest_framework import exceptions


def check_user_validity(user):
    """
    Check validity of given user.

    :type user: django.contrib.auth.models.AbstractBaseUser
    :rtype: None
    """
    if not user.is_active:
        raise exceptions.PermissionDenied()


def _unpadded_encode(data, encoder=base64.b64encode):
    return encoder(data).rstrip(b'=').decode('ascii')


def hash_string(string, hasher=hashlib.sha256, formatter=_unpadded_encode):
    """
    Calculate a hash of given string encoded to UTF-8.

    The hasher is SHA-256 is by default, but it can be changed with an
    argument.  T binary hash returned by the hasher is converted to a
    string with the given formatter; the default formatter returns the
    hash as unpadded base 64 string.

    :type string: str
    :type hasher: Callable[[bytes], bytes]
    :type formatter: Callable[[bytes], str]
    :rtype: str
    """
    return formatter(hasher(string.encode('utf-8')).digest())


def sha1_string(string, formatter=_unpadded_encode):
    return hash_string(string, hasher=hashlib.sha1, formatter=formatter)



from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions, status


class VerificationCodeSendingFailed(exceptions.APIException):
    status_code = status.HTTP_501_NOT_IMPLEMENTED
    default_code = 'verification_code_sending_failed'
    default_detail = _("Verification code sending failed: {reason}")

    def __init__(self, reason, detail=None, code=None, *args, **kwargs):
        super(VerificationCodeSendingFailed, self).__init__(
            detail=(detail or self.default_detail.format(reason=reason)),
            code=code, *args, **kwargs)
